#define THREAD_USE_POSIX
#include "libtsctl.h"
#include "Array.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>

int addhost(char **hosts,char *host,int found) {
  int i;
  for (i=0;i<found;i++) {
    if (!memcmp(hosts[i],host,6)) return 0;
  }
  memcpy(hosts[i],host,6);
  return 1;
}

/*
Ok, there are two different types of messages to send/receive

1. HELLO
id=0x3110
len=6
data=mac address

2. MESSAGE
id=0x0414XXAB
len=(A+B) mod 8
data=rand() * len with srand(XXAB)
 */
int done=0;
CAN *can;
static void alarmsig(int x) {
  can->Abort(can);
  done = 1;
}

int main(int argc,char *argv[]) {
  int i,j,k,len,n,rc,found=0,err=0;
  FILE *f;
  int mac[6];
  CANMessage msg;
  ArrayAutoOfSize(char,m,8);
  char **hosts,byte;
  int *rx,rxt=0,tx=0,rxexpect;

  if (argc < 2) {
    printf("specify total number of boards running test\n");
    return 1;
  }
  n = atoi(argv[1]);
  if (n < 2) {
    printf("must specify at least 2 boards running test\n");
    return 2;
  }
  hosts = malloc(sizeof(char *) * n);
  for (i=0;i<n;i++) {
    hosts[i] = malloc(6);
  }
  rx = malloc(sizeof(int) * n);
  for (i=0;i<n;i++) rx[i] = 0;
  can = CANInit(0);
  if (!can) {
    fprintf(stderr,"Error: No CAN present\n");
    return 1;
  }
  if (SignalCapture(2,alarmsig) < 0) {
    fprintf(stderr,"signal capture error:%m\n");
    return 1;
  }
  if (SignalCapture(14,alarmsig) < 0) {
    fprintf(stderr,"signal capture error:%m\n");
    return 1;
  }

  f = fopen("/sys/class/net/eth0/address","r");
  fscanf(f,"%x:%x:%x:%x:%x:%x",mac,mac+1,mac+2,mac+3,mac+4,mac+5);
  fclose(f);
  for (i=0;i<6;i++) {
    m[i] = mac[i];
    msg.data[i] = 0x55;
    //printf("%02X ",m[i]);
  }
  //printf("\n");
  ArraySizeAuto(m,6);
  can->BaudSet(can,1000000);
  can->Tx(can,FLAG_EXT_ID,0x3110,m);
  alarm(5);
  for (found=0;found<n-1&&!done;) {
    rc = can->Rx(can,&msg);
    if (rc > 0) {
      if (msg.id != 0x3110 || !(msg.flags & FLAG_EXT_ID)
	  || ArrayLength(msg.data) != 6) {
	//printf("id=%X\n",msg.id);
      } else {
	if (addhost(hosts,msg.data,found)) {
	  printf("found:");
	  for (j=0;j<6;j++) {
	    printf("%02X ",msg.data[j]);
	  }
	  printf("\n");
	  found++;
	} else printf("skipped duplicate\n");
      }
    } //else printf("rc=%d\n",rc);
  }
  alarm(0); // cancel alarm
  //printf("found a total of %d\n",found);
  if (found < n-1) {
    printf("failed to find enough boards\n");
    return 1;
  }
  rxexpect = 10 * (n-1);

  while (!done && (tx < 10 || rxt < rxexpect)) {
    if (tx < 10) {
      // note: following code is incorrect if i > 255
      len = 1 + tx % 7;
      srand(tx);
      for (j=0;j<len;j++) {
	m[j] = (char)rand();
      }
      ArraySizeAuto(m,len);
      fprintf(stderr,"?%d",tx);
      can->Tx(can,FLAG_EXT_ID,0x04140000+tx,m);
      tx++;
    }
    if (rxt < rxexpect) {
      for (k=0;!done && k<n-1 && rxt < rxexpect;k++) {
	alarm(1);
	rc = can->Rx(can,&msg);
	fprintf(stderr,".");
	alarm(0);
	if (!done) {
	  if (rc > 0) {
	    if ((msg.id>>16)==0x0414) {
	      rxt++;
	      len = 1 + (msg.id & 255) % 7;
	      if (msg.length != len) {
		fprintf(stderr,"len=%d, expected %d\n",msg.length,len);
		err++;
	      } else {
		fprintf(stderr,"%d",msg.id & 0xFFFF);
		srand(msg.id & 0xFFFF);
		for (i=0;i<len;i++) {
		  byte = (char)rand();
		  if (byte != msg.data[i]) {
		    fprintf(stderr,"[%d]=%X, expected %X msg.id=%X\n",i,msg.data[i],byte,msg.id);
		    err++;
		    break;
		  }
		}
	      }
	    } //else fprintf(stderr,"<%X>",msg.id);
	  } else { fprintf(stderr,"rc=%d\n",rc); done = 1; }
	} else {
	  fprintf(stderr,"timeout (done=%d)\n",done);
	}
      }
    }
    printf("done\n");
  }
  return err>99 ? 99 : 0;
}
